import typing

from graphql import GraphQLUnionType

from .exceptions import UnallowedReturnTypeForUnion, WrongReturnTypeForUnion
from .utils.typing import get_list_annotation, is_generic, is_list, is_type_var


def _get_type_mapping_from_actual_type(root) -> typing.Dict[typing.Any, typing.Type]:
    # we map ~T to the actual type of root
    type_var_to_actual_type = {}

    for field_name, annotation in root.__annotations__.items():
        # when we have a list we want to get the type of the elements contained in the
        # list, to do so we currently only get the first time (if the list is not empty)
        # this might break in more complex cases, but should suffice for now.

        if is_list(annotation):
            annotation = get_list_annotation(annotation)

            if is_type_var(annotation):
                values = getattr(root, field_name)

                if values:
                    type_var_to_actual_type[annotation] = type(values[0])

        elif is_type_var(annotation):
            type_var_to_actual_type[annotation] = type(getattr(root, field_name))

        elif is_generic(annotation):
            type_var_to_actual_type.update(
                _get_type_mapping_from_actual_type(getattr(root, field_name))
            )

    return type_var_to_actual_type


def _find_type_for_generic_union(root: typing.Any) -> typing.Type:
    # this is a ordered tuple of the type vars for the generic class, so for
    # typing.Generic[T, V] it would return (T, V)
    type_params = root.__parameters__

    mapping = _get_type_mapping_from_actual_type(root)

    if not mapping:
        # if we weren't able to find a mapping, ie. when returning an empty list
        # for a generic type, then we fall back to returning the first copy.
        # This a very simplistic heuristic and it is bound to break with complex
        # uses cases. We can improve it later if this becomes an issue.

        return next((t for t in root._copies.values()))

    types = tuple(mapping[param] for param in type_params)

    type = root._copies.get(types)

    if type is None:
        raise ValueError(f"Unable to find type for {root.__class__} and {types}")

    return type


def union(name: str, types: typing.Tuple[typing.Type], *, description=None):
    """Creates a new named Union type.

    Example usages:

    >>> strawberry.union(
    >>>     "Name",
    >>>     (A, B),
    >>> )

    >>> strawberry.union(
    >>>     "Name",
    >>>     (A, B),
    >>> )
    """

    from .type_converter import get_graphql_type_for_annotation

    def _resolve_type(root, info, _type):
        if not hasattr(root, "graphql_type"):
            raise WrongReturnTypeForUnion(info.field_name, str(type(root)))

        if is_generic(type(root)):
            return _find_type_for_generic_union(root)

        if root.graphql_type not in _type.types:
            raise UnallowedReturnTypeForUnion(
                info.field_name, str(type(root)), _type.types
            )

        return root.graphql_type

    # TODO: union types don't work with scalar types
    # so we want to return a nice error
    # also we want to make sure we have been passed
    # strawberry types
    graphql_type = GraphQLUnionType(
        name,
        [
            get_graphql_type_for_annotation(type, name, force_optional=True)
            for type in types
        ],
        description=description,
    )
    graphql_type.resolve_type = _resolve_type

    # This is currently a temporary solution, this is ok for now
    # But in future we might want to change this so that it works
    # properly with mypy, but there's no way to return a type like NewType does
    # so we return this class instance as it allows us to reuse the rest of
    # our code without doing too many changes

    class X:
        def __init__(self, graphql_type):
            self.graphql_type = graphql_type

    return X(graphql_type)



import dataclasses
from itertools import chain
from typing import Callable

from graphql.type import (
    GraphQLDirective,
    GraphQLNamedType,
    GraphQLSchema,
    is_object_type,
    is_specified_directive,
)
from graphql.utilities.print_schema import (
    is_defined_type,
    print_args,
    print_block,
    print_deprecated,
    print_description,
    print_directive,
    print_implemented_interfaces,
    print_schema_definition,
    print_type as original_print_type,
)

from .type_registry import get_strawberry_type_for_graphql_type


def print_federation_field_directive(field, metadata):
    out = ""

    if metadata and "federation" in metadata:
        federation = metadata["federation"]

        provides = federation.get("provides", "")
        requires = federation.get("requires", "")
        external = federation.get("external", False)

        if provides:
            out += f' @provides(fields: "{provides}")'

        if requires:
            out += f' @requires(fields: "{requires}")'

        if external:
            out += " @external"

    return out


def print_fields(type_) -> str:
    strawberry_type = get_strawberry_type_for_graphql_type(type_)
    strawberry_fields = dataclasses.fields(strawberry_type) if strawberry_type else []

    def _get_metadata(field_name):
        return next(
            (
                f.metadata
                for f in strawberry_fields
                if (getattr(f, "field_name", None) or f.name) == field_name
            ),
            None,
        )

    fields = [
        print_description(field, "  ", not i)
        + f"  {name}"
        + print_args(field.args, "  ")
        + f": {field.type}"
        + print_federation_field_directive(field, _get_metadata(name))
        + print_deprecated(field)
        for i, (name, field) in enumerate(type_.fields.items())
    ]
    return print_block(fields)


def print_federation_key_directive(type_):
    strawberry_type = get_strawberry_type_for_graphql_type(type_)

    if not strawberry_type:
        return ""

    keys = getattr(strawberry_type, "_federation_keys", [])

    parts = []

    for key in keys:
        parts.append(f'@key(fields: "{key}")')

    if not parts:
        return ""

    return " " + " ".join(parts)


def print_extends(type_):
    strawberry_type = get_strawberry_type_for_graphql_type(type_)

    if strawberry_type and getattr(strawberry_type, "_federation_extend", False):
        return "extend "

    return ""


def print_object(type_) -> str:
    return (
        print_description(type_)
        + print_extends(type_)
        + f"type {type_.name}"
        + print_federation_key_directive(type_)
        + print_implemented_interfaces(type_)
        + print_fields(type_)
    )


def print_type(field) -> str:
    """Returns a string representation of a strawberry type"""

    if hasattr(field, "graphql_type"):
        field = field.graphql_type

    if is_object_type(field):
        return print_object(field)

    return original_print_type(field)


def print_filtered_schema(
    schema: GraphQLSchema,
    directive_filter: Callable[[GraphQLDirective], bool],
    type_filter: Callable[[GraphQLNamedType], bool],
) -> str:
    directives = filter(directive_filter, schema.directives)
    type_map = schema.type_map

    types = filter(type_filter, map(type_map.get, sorted(type_map)))  # type: ignore

    return "\n\n".join(
        chain(
            filter(None, [print_schema_definition(schema)]),
            (print_directive(directive) for directive in directives),
            (print_type(type_) for type_ in types),  # type: ignore
        )
    )


def print_schema(schema: GraphQLSchema) -> str:
    return print_filtered_schema(
        schema, lambda n: not is_specified_directive(n), is_defined_type
    )



import typing

from graphql import (
    GraphQLBoolean,
    GraphQLFloat,
    GraphQLID,
    GraphQLInt,
    GraphQLString,
    GraphQLType,
)

from .scalars import ID


TYPE_HINTS_REGISTRY: typing.Dict[str, typing.Type] = {}
ANNOTATION_REGISTRY: typing.Dict[typing.Type, GraphQLType] = {
    str: GraphQLString,
    int: GraphQLInt,
    float: GraphQLFloat,
    bool: GraphQLBoolean,
    ID: GraphQLID,
}

GRAPHQL_TYPE_TO_STRAWBERRY_TYPE: typing.Dict[GraphQLType, typing.Type] = {}


def register_type(
    cls: typing.Type, graphql_type: GraphQLType, *, store_type_information: bool = True
) -> None:
    """Register a Strawberry type alongside its GraphQL type"""

    # register the annotation so we can use when getting the graphql type for
    # an annotation in type_converter
    ANNOTATION_REGISTRY[cls] = graphql_type
    GRAPHQL_TYPE_TO_STRAWBERRY_TYPE[graphql_type] = cls

    # register names to classes so that we can pass them to get_type_hints
    TYPE_HINTS_REGISTRY[cls.__name__] = cls

    # we currently have to store the "graphql core" type onto the
    # class, but we can't do that for all types (like builtins),
    # so we have an escape hatch for that

    if store_type_information:
        cls.graphql_type = graphql_type


def get_strawberry_type_for_graphql_type(
    graphql_type: GraphQLType,
) -> typing.Optional[typing.Type]:
    return GRAPHQL_TYPE_TO_STRAWBERRY_TYPE.get(graphql_type)


def get_type_for_annotation(annotation: typing.Type) -> typing.Optional[GraphQLType]:
    return ANNOTATION_REGISTRY.get(annotation)


def get_registered_types() -> typing.Dict[str, typing.Type]:
    return TYPE_HINTS_REGISTRY



from graphql import (
    GraphQLField,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
)

from .field import strawberry_field
from .printer import print_schema
from .schema import Schema as BaseSchema
from .type import _process_type
from .type_registry import get_strawberry_type_for_graphql_type


def type(cls=None, *args, **kwargs):
    def wrap(cls):
        keys = kwargs.pop("keys", [])
        extend = kwargs.pop("extend", False)

        wrapped = _process_type(cls, *args, **kwargs)
        wrapped._federation_keys = keys
        wrapped._federation_extend = extend

        return wrapped

    if cls is None:
        return wrap

    return wrap(cls)


def strawberry_federation_field(*args, **kwargs):
    provides = kwargs.pop("provides", "")
    requires = kwargs.pop("requires", "")
    external = kwargs.pop("external", False)

    metadata = kwargs.get("metadata") or {}
    metadata["federation"] = {
        "provides": provides,
        "external": external,
        "requires": requires,
    }
    kwargs["metadata"] = metadata

    field = strawberry_field(*args, **kwargs)

    return field


def field(wrap=None, *args, **kwargs):
    field = strawberry_federation_field(*args, **kwargs)

    if wrap is None:
        return field

    return field(wrap)


def entities_resolver(root, info, representations):
    results = []

    for representation in representations:
        type_name = representation.pop("__typename")
        graphql_type = info.schema.get_type(type_name)

        result = get_strawberry_type_for_graphql_type(graphql_type).resolve_reference(
            **representation
        )
        results.append(result)

    return results


def has_federation_keys(graphql_type: GraphQLType):
    strawberry_type = get_strawberry_type_for_graphql_type(graphql_type)

    if strawberry_type and getattr(strawberry_type, "_federation_keys", []):
        return True

    return False


class Schema(BaseSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._extend_query_type()

    def _extend_query_type(self):
        @type(name="_Service")
        class Service:
            sdl: str

        Any = GraphQLScalarType("_Any")

        fields = {
            "_service": GraphQLField(
                GraphQLNonNull(Service.graphql_type),
                resolve=lambda _, info: Service(sdl=print_schema(info.schema)),
            )
        }

        entities_type = self._get_entity_type()

        if entities_type:
            self.type_map[entities_type.name] = entities_type

            fields["_entities"] = GraphQLField(
                GraphQLNonNull(GraphQLList(entities_type)),
                args={
                    "representations": GraphQLNonNull(GraphQLList(GraphQLNonNull(Any)))
                },
                resolve=entities_resolver,
            )

        fields.update(self.query_type.fields)

        self.query_type = GraphQLObjectType(
            name=self.query_type.name,
            description=self.query_type.description,
            fields=fields,
        )

        self.type_map["_Any"] = Any
        self.type_map["_Service"] = Service.graphql_type
        self.type_map[self.query_type.name] = self.query_type

    def _get_entity_type(self):
        # https://www.apollographql.com/docs/apollo-server/federation/federation-spec/#resolve-requests-for-entities

        # To implement the _Entity union, each type annotated with @key
        # should be added to the _Entity union.

        federation_key_types = [
            graphql_type
            for graphql_type in self.type_map.values()
            if has_federation_keys(graphql_type)
        ]

        # If no types are annotated with the key directive, then the _Entity
        # union and Query._entities field should be removed from the schema.
        if not federation_key_types:
            return None

        entity_type = GraphQLUnionType("_Entity", federation_key_types)

        def _resolve_type(self, value, _type):
            return self.graphql_type

        entity_type.resolve_type = _resolve_type

        return entity_type
